use galactica_content::{Content, FactionHandle, OutfitHandle, ShipHandle, SystemHandle};
use galactica_galaxy::{ship::ShipPersonality, Galaxy, GxShipHandle};
use galactica_playeragent::{PlayerAgent, PlayerStatus};
use galactica_render::RenderInput;
use galactica_systemsim::{util, ParticleBuilder, StepResources, SystemSim, Wrapper};
use galactica_util::timing::Timing;
use std::time::Instant;

#[derive(Clone)]
pub struct GameState {
	pub gx: Galaxy,
	pub systemsim: SystemSim,
}

pub struct Game {
	// Core game data
	ct: Content,
	state_live: GameState,
	state_static: GameState,

	// Metadata
	wrapper: Wrapper, // Physics computer
	time_scale: f32,
	last_update: Instant,
	start_instant: Instant,
	new_particles: Vec<ParticleBuilder>,
	timing: Timing,
}

impl Game {
	pub fn new(ct: Content) -> (Self, GxShipHandle) {
		let mut gx = Galaxy::new(&ct);

		let player = gx.create_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 0 },
			ShipPersonality::Player,
			&SystemHandle { index: 0 },
		);
		let s = gx.get_ship_mut(player).unwrap();
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 0 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let a = gx.create_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 1 },
			ShipPersonality::Dummy,
			&SystemHandle { index: 0 },
		);
		let s = gx.get_ship_mut(a).unwrap();
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 0 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let a = gx.create_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 0 },
			ShipPersonality::Point,
			&SystemHandle { index: 0 },
		);
		let s = gx.get_ship_mut(a).unwrap();
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 0 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let systemsim = SystemSim::new(&ct, &gx, SystemHandle { index: 0 });

		let state = GameState { gx, systemsim };

		(
			Game {
				ct,
				state_live: state.clone(),
				state_static: state,

				wrapper: Wrapper::new(),

				last_update: Instant::now(),
				start_instant: Instant::now(),
				time_scale: 1.0,
				new_particles: Vec::new(),
				timing: Timing::new(),
			},
			player,
		)
	}

	pub fn update_player_controls(&mut self, player: &PlayerAgent) {
		self.state_live.systemsim.update_player_controls(player)
	}

	pub fn get_player_status(&self, player: &PlayerAgent) -> PlayerStatus {
		let o = self
			.state_static
			.systemsim
			.get_ship(player.ship.unwrap())
			.unwrap();
		let r = self
			.state_static
			.systemsim
			.get_rigid_body(o.rigid_body)
			.unwrap();
		let pos = util::rigidbody_position(r);

		PlayerStatus { pos }
	}

	pub fn update(&mut self) {
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;

		self.timing.start_frame();
		self.timing.start_galaxy();
		self.state_live.gx.step(t);
		self.timing.mark_galaxy();

		self.state_live.systemsim.step(StepResources {
			ct: &self.ct,
			gx: &mut self.state_live.gx,
			particles: &mut self.new_particles,
			timing: &mut self.timing,
			wrapper: &mut self.wrapper,
			t,
		});

		self.last_update = Instant::now();
		self.state_static = self.state_live.clone();
		self.timing.mark_frame();
	}

	pub fn get_frame_state(&mut self, player: &PlayerAgent) -> RenderInput {
		RenderInput {
			camera_pos: player.camera.pos,
			camera_zoom: player.camera.zoom,
			current_time: self.start_instant.elapsed().as_secs_f32(),
			ct: &self.ct,
			systemsim: &self.state_static.systemsim, // TODO: system should be stored here, not in game
			particles: &mut self.new_particles,
			player_data: player.ship.unwrap(),
			gx: &self.state_static.gx,
			current_system: SystemHandle { index: 0 },
			timing: &mut self.timing,
		}
	}

	pub fn get_content(&self) -> &Content {
		&self.ct
	}
}
